#! /usr/bin/python3

from scipy.signal import butter, lfilter
from scipy.signal import freqz
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

def plot(y, ylabel, x, xlabel, title):
    fig, ax = plt.subplots()
    ax.plot(x, y)
    ax.set(xlabel=xlabel, ylabel=ylabel, title=title)
    ax.grid()

# Code taken from stackoverflow
# https://stackoverflow.com/questions/12093594/how-to-implement-band-pass-butterworth-filter-with-scipy-signal-butter
def butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a

def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y

def bin2int(binary):
    idx=0
    ret=0
    for bit in binary:
        ret = ret | (bit << idx)
        idx = idx + 1
    return ret

# Import measures
consumption = np.load("./arr_0.npy")

# RSA start at sample 700000 and ends around 1490000
#consumption = consumption[700000:1490000]

# Count samples for plotting
# Compute as 1Hz sample, we don't really care
t = np.arange(len(consumption))

plot(consumption, 'consumption', t, 'time', 'Consumption')

consumption = consumption[700000:1490000]
t = np.arange(len(consumption))

# FFT computation
n = len(consumption) # length of the signal
k = np.arange(n)
frq = k/n # two sides frequency range
frq = frq[range(int(n/2))] # one side frequency range
Y = np.fft.fft(consumption)/n # fft computing and normalization
Y = Y[range(int(n/2))]

# plot(Y, 'amplitude', frq, 'frequency', 'FFT')
# FFT show that most signal frequency is before 0.005 but interresting peaks for RSA "square" and "multiply"
# are very thin, means there are on high frequency. So I used a filter to get only high freq.

# Signal filtering
# We don't really need a bandpass filter as we just want to drop low freq
# Also, note that freq are multiple of sampling frquency (which I set to 1Hz as we don't know it)
# So 0.01 means we drop everyhing under 1% of the sampling freq. We also drop very high freq (from 40% to 50%).
consumption = butter_bandpass_filter(consumption, 0.01, 0.4, 1.0, order=1)
#plot(consumption, 'consumption', t, 'time', 'Consumption')
#plt.show()

# Here we just count number of peaks during one iteration of the fast exp. modul.
# Two peaks means this is just "square" -> Di = 0
# Four peaks means this is "square and multiply" -> Di = 1
#
# Note that negative peaks are more usable, so we use a negative threshold
threshold = -0.20
# Maximum time beetween two peak to understand that is a new iteration (D[i+1])
max_time_between_2peak = 250

D = list()
peaks = 0
below_threshold = False
samples_without_peaks = 0

for sample in consumption:
    # Increment peaks each time we see one
    if sample < threshold:
        samples_without_peaks = 0
        # New peak seen
        if not below_threshold:
            below_threshold = True
            peaks = peaks + 1
    else:
        samples_without_peaks = samples_without_peaks + 1
        below_threshold = False

    # Store peaks seen between each fast exp. mod. iteration
    if peaks > 0 and samples_without_peaks > max_time_between_2peak:
        if peaks == 4:
            D.append(1)
        elif peaks == 2:
            D.append(0)
        else:
            print('Error: wrong number of peaks seen')
        peaks = 0

print('Len of key is %d' % len(D))
print(repr(D))

hexakey = ''
for i in range(int(len(D)/8)):
    print('Conversion of %s is %02X' % (repr(list(D[i*8:(i+1)*8])), bin2int(D[i*8:(i+1)*8])))
    hexakey = '%02X' % bin2int(D[i*8:(i+1)*8]) + hexakey

print('--------------------------------')
print('Key is ' + hexakey)

# Key is :
# 23D87CDF97BB95ABE6273C384190C765F552AB86F6DE30A8DB74435C95E6E3138F54AF689812D8F9359CF0F4D453A0C11EC68CE470216C09E74C8947ADAF23E902415D61DDF2C0FFE459CBB40F7DE42BDB7CD14093100A570E8C29819765E2D8D276F86471B52AC29AA2CE2BB72CD45006279E82BEC253AE9675FE45824F6001
