#!/usr/bin/python
# -*- coding: utf-8 -*

import hashlib
import sys
import binascii
import random
from Crypto.Cipher import AES

def mix_input(inpt, btnX):
	if not btnX:
		return inpt
	else:
		return ((inpt << 1) & 0xff) | ((inpt & 0x80) >> 7)

def mix_op(op, btn1, btn2):
	return ((op & 1) ^ btn1) | ((op & 2) ^ (btn2 << 1))

def mix_out(a, b, op):
	mask=0
	outTmpPrev=0

	if op == 0:
		return (a & b)
	if op == 1:
		return (a | b)
	if op == 2:
		return (a ^ b)
	# Otherwise op == 3
	for i in range(8):
		maskn = (1 << i)
		an = (a & maskn)
		bn = (b & maskn)
		outTmpPrev = (outTmpPrev << 1)
		mask = mask | outTmpPrev
		outTmpPrev = (outTmpPrev & (an ^ bn)) | (an & bn)
	return (a ^ b ^ mask)

def secure_device(a,b,op):
	out = 0

	btn1 = 1 if (btn & 1) else 0
	btn2 = 1 if (btn & 2) else 0
	btn3 = 1 if (btn & 4) else 0
	btn4 = 1 if (btn & 8) else 0

	#print("Input a=%02x b=%02x op=%02x" % (a, b, op))
	op = mix_op(op, btn1, btn2)
	a = mix_input(a, btn3)
	b = mix_input(b, btn4)
	#print("Mixed a=%02x b=%02x op=%02x" % (a, b, op))
	out = mix_out(a, b, op)
	#print("Out = %02x" % (out))

	return out

def step1():
	r = secure_device(0x35,0x27,3)
	r = secure_device(0x7e,r,3)
	r = secure_device(0x66,r,2)
	r = secure_device(0x8,r,1)
	r = secure_device(0x13,r,0)
	r = secure_device(0x1f,r,1)
	r = secure_device(0xa,r,2)
	r = secure_device(0xd3,r,0)
	r = secure_device(0xc6,r,3)

	return r

def step2():
	r= secure_device(0xde,0xab,0)
	r= secure_device(0x67,r,3)
	r= secure_device(0x2a,r,2)
	r= secure_device(0x6d,r,1)
	r= secure_device(0x4a,r,3)
	r= secure_device(0xe7,r,0)
	r= secure_device(0x1c,r,1)
	r= secure_device(0x35,r,0)
	r= secure_device(0xde,r,3)
	r= secure_device(0xf7,r,0)
	r= secure_device(0xda,r,2)
	return r

def step3():
	r = secure_device(0x14,0x23,3)
	r = secure_device(0x72,r,0)
	r = secure_device(0x48,r,3)
	r = secure_device(0x53,r,1)
	r = secure_device(0xa7,r,0)
	r = secure_device(0x5f,r,1)
	r = secure_device(0x3,r,3)
	r = secure_device(0xb7,r,3)
	r = secure_device(0x73,r,1)
	r = secure_device(0x37,r,3)
	r = secure_device(0xc5,r,2)
	r = secure_device(0xa4,r,1)
	r = secure_device(0x30,r,0)
	r = secure_device(0xdd,r,2)
	return r


def step4():
	r = secure_device(0xb0,0x42,2)
	r = secure_device(0xbc,r,2)
	r = secure_device(0xfc,r,2)
	r = secure_device(0x54,r,3)
	r = secure_device(0x30,r,2)
	r = secure_device(0x97,r,1)
	r = secure_device(0xe8,r,2)
	r = secure_device(0xd6,r,0)
	r = secure_device(0x26,r,0)
	r = secure_device(0xeb,r,0)
	r = secure_device(0x68,r,1)
	r = secure_device(0x26,r,0)
	r = secure_device(0x9,r,3)
	r = secure_device(0x2a,r,2)
	r = secure_device(0xa9,r,3)
	return r


def step5():
	r = secure_device(0xff,0x12,0)
	r = secure_device(0xfd,r,1)
	r = secure_device(0xe5,r,1)
	r = secure_device(0x26,r,3)
	r = secure_device(0x85,r,3)
	r = secure_device(0x63,r,1)
	r = secure_device(0x93,r,3)
	r = secure_device(0xba,r,2)
	r = secure_device(0x97,r,0)
	r = secure_device(0xab,r,1)
	r = secure_device(0x6e,r,3)
	r = secure_device(0xfd,r,0)
	r = secure_device(0x4c,r,3)
	r = secure_device(0x50,r,0)
	r = secure_device(0xa,r,2)
	r = secure_device(0xfc,r,3)
	r = secure_device(0xe3,r,2)
	r = secure_device(0xa6,r,3)
	r = secure_device(0x64,r,2)
	r = secure_device(0x8e,r,3)
	r = secure_device(0xc1,r,1)
	return r

def step6():
	r = secure_device(0x90,0x77,1)
	r = secure_device(0x8e,r,0)
	r = secure_device(0xbd,r,2)
	r = secure_device(0x39,r,2)
	r = secure_device(0x4c,r,2)
	r = secure_device(0xc5,r,2)
	r = secure_device(0xb6,r,3)
	r = secure_device(0x93,r,1)
	r = secure_device(0x9f,r,3)
	r = secure_device(0xd6,r,3)
	r = secure_device(0x6e,r,2)
	r = secure_device(0x39,r,3)
	r = secure_device(0x40,r,1)
	r = secure_device(0x14,r,2)
	r = secure_device(0xe6,r,3)
	return r

def step7():
	r = secure_device(0xf,0xab,3)
	r = secure_device(0xa2,r,1)
	r = secure_device(0x7c,r,0)
	r = secure_device(0x34,r,1)
	r = secure_device(0x14,r,1)
	r = secure_device(0xe7,r,0)
	r = secure_device(0xb9,r,0)
	r = secure_device(0xf1,r,2)
	r = secure_device(0xd5,r,1)
	r = secure_device(0x4e,r,2)
	r = secure_device(0xe,r,2)
	r = secure_device(0x6,r,0)
	r = secure_device(0x7d,r,2)
	r = secure_device(0x87,r,3)
	r = secure_device(0xbc,r,0)
	r = secure_device(0xd4,r,3)
	r = secure_device(0x8a,r,1)
	r = secure_device(0xe7,r,3)
	r = secure_device(0x9e,r,1)
	r = secure_device(0x58,r,0)
	r = secure_device(0x24,r,2)
	r = secure_device(0x44,r,3)
	r = secure_device(0xc9,r,1)
	r = secure_device(0xd4,r,1)
	r = secure_device(0x1d,r,3)
	r = secure_device(0xcd,r,0)
	r = secure_device(0xde,r,1)
	r = secure_device(0x54,r,0)
	r = secure_device(0x5e,r,2)
	r = secure_device(0x46,r,1)
	r = secure_device(0x21,r,0)
	r = secure_device(0xff,r,1)
	r = secure_device(0x51,r,0)
	r = secure_device(0x78,r,1)
	r = secure_device(0x2f,r,3)
	r = secure_device(0xed,r,2)
	r = secure_device(0x4b,r,3)
	r = secure_device(0x4d,r,2)
	return r

def step8():
	r = secure_device(0x88,0x74,0)
	r = secure_device(0x48,r,2)
	r = secure_device(0x11,r,2)
	r = secure_device(0x76,r,0)
	r = secure_device(0x2b,r,3)
	r = secure_device(0xf8,r,2)
	return r

def init():
	r = secure_device(0x46,0x92,0)
	r = secure_device(0xdf,r,2)
	r = secure_device(0x3e,r,0)
	r = secure_device(0x3a,r,3)
	r = secure_device(0x36,r,2)
	r = secure_device(0x8e,r,2)
	r = secure_device(0xc9,r,3)
	r = secure_device(0xe7,r,1)
	r = secure_device(0x29,r,2)
	r = secure_device(0xc2,r,2)
	r = secure_device(0x79,r,0)
	r = secure_device(0x2a,r,2)
	r = secure_device(0x4c,r,3)
	r = secure_device(0xde,r,0)
	r = secure_device(0x88,r,0)
	r = secure_device(0x8b,r,2)
	r = secure_device(0x97,r,3)
	r = secure_device(0x6a,r,2)
	r = secure_device(0x60,r,1)
	r = secure_device(0x0f,r,0)
	r = secure_device(0x5b,r,3)
	r = secure_device(0xd0,r,2)
	r = secure_device(0xa9,r,1)
	r = secure_device(0xe3,r,3)
	r = secure_device(0xd0,r,1)
	r = secure_device(0x27,r,0)
	r = secure_device(0x90,r,0)
	r = secure_device(0x3b,r,1)
	r = secure_device(0x66,r,2)
	r = secure_device(0xe2,r,0)
	r = secure_device(0x24,r,3)
	r = secure_device(0xee,r,1)
	r = secure_device(0xf2,r,3)
	return r

def main():
	global btn
	for x in range(0xf+1):
		btn = x
		s = step1()
		print('s1: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step2()
		print('s2: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step3()
		print('s3: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step4()
		print('s4: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step5()
		print('s5: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step6()
		print('s6: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step7()
		print('s7: %02x' % s)
	for x in range(0xf+1):
		btn = x
		s = step8()
		print('s8: %02x' % s)

if __name__ == '__main__':
	main()
