#! /usr/bin/python

# Special register is reg[0xf] which is PC
reg = dict()
instr = dict()
lookup_table = dict()
global_return = -1

# Constants hard-coded in the EL3 code
constants = [ 0x612e7270, 0x6766722e, 0x666e632e, 0x2e76662e,
              0x76706e73, 0x66407279, 0x70766766, 0x7465622e ]

# --------------------------------------------
#   SM4 encrypted data manipulation
#   These operation are handled using
#   pagefaults.
# --------------------------------------------
# Note that I already decrypted static data like
# VM code and lookup table
# 00000000 : VM Code
# 00001000 : Lookup table
# 00100000 : Runtime storage

def runtime_storage_add(addr, val32b):
	# TODO
	return

def lookup_reverse(val32b):
	for val1 in lookup_table:
		if lookup_table[val2] == val32b:
			return val1
	# print "Lookup error"
	return 0

def lookup(val32b):
	return lookup_table[val32b]

#
# --------------------------------------------
#   Instructions implementation
# --------------------------------------------
# Instruction are coded on 24 bits (0xffffff)
# So addr are increased by 3.
# Structure is like :
#     | OP1  | OP2  | VAL1 | VAL2 |      |
# or  | OP1  | OP2  | VAL1 |     VAL4    |
#     24     20     18     14            0
#
# The actual operation depends on OP1 and OP2
# I could have mixed them into one OP but it's easier for me :)
def process_one(addr, instr):
	global reg
	op1 = (instr >> 20) & 0xf
	op2 = (instr >> 18) & 0x3
	val1 = (instr >> 14) & 0xf
	val2 = (instr >> 10) & 0xf
	val4 = instr & 0x3fff

	print "0x%04x : %06x" % (addr, instr),

	if op1 == 0xa :
		op_a_x(val2)
		return
	if op1 == 0xc :
		op_c_x(val1)
		return
	if op1 == 0xd :
		op_d_x()
		return
	if op1 == 0xe or op1 == 0xf:
		op_ef_x(val4)
		return
	if op2 == 0x1 :
		op_x_1(val1, val2)
		return
	if op2 == 0x2 :
		op_x_2(val1, val2)
		return
	if op2 == 0x3 :
		if op1 == 0:
			op_0_3(val1, val4)
			return
		if op1 == 2:
			op_2_3(val1, val4)
			return
		if op1 == 3:
			op_3_3(val1, val4)
			return
		if op1 == 4:
			op_4_3(val1, val4)
			return
		if op1 == 5:
			op_5_3(val1, val4)
			return
		if op1 == 7:
			op_7_3(val1, val4)
			return
		if op1 == 8:
			op_8_3(val4)
			return
		if op1 == 9:
			op_9_3(val1, val4)
			return
	if op2 == 0x0 :
		if op1 == 0:
			op_0_0(val1, val2)
			return
		if op1 == 1:
			op_1_0(val1)
			return
		if op1 == 2:
			op_2_0(val1, val2)
			return
		if op1 == 3:
			op_3_0(val1, val2)
			return
		if op1 == 6:
			op_6_0(val1, val2)
			return
		if op1 == 0xb:
			op_b_0(val1)
			return

	print "OPCODE : invalid"
	reg[0xf] += 3

# Function are called like : op_$op1_$op2()
# X means any value.

def op_c_x(val1):
	global reg
	index = reg[val1] & 0xfc
	addr = 0x100000 + index
	print "OPCODE : storage[reg[0x%01x]] = constants[reg[0x%01x]]" % (val1,val1)
	#print "STORAGE : change index %d with value %08x)" % (addr, constants[index/4])
	runtime_storage_add(addr, constants[index/4])
	reg[0xf] += 3

def op_a_x(val2):
	global global_return
	global reg
	print "OPCODE : return final value from reg[0x%01x]" % val2
	global_return = reg[val2]
	reg[0xf] += 3

# Anti debug
def op_d_x():
	global reg
	#print "OPCODE : *0x09010008 = 0"
	print "OPCODE : nop"
	reg[0xf] += 3

# Anti debug
def op_ef_x(val4):
	global reg
	#print "OPCODE : goto 0x%04x if *0x09010000 <= 5" % (val4)
	print "OPCODE : nop"
	reg[0xf] += 3

def op_x_1(val1, val2):
	global reg
	print "OPCODE : reg[0x%01x] = lookup(reg[0x%01x])" % (val1, val2)
	reg[val1] = lookup(reg[val2])
	reg[0xf] += 3

def op_x_2(val1, val2):
	global reg
	print "OPCODE : storage[reg[0x%01x]] = reg[0x%01x]" % (val1,val2)
	#print "STORAGE : change index %d with value %08x)" % (reg[val1], reg[val2])
	runtime_storage_add(reg[val1], reg[val2])
	reg[0xf] += 3

def op_0_3(val1, val4):
	global reg
	print "OPCODE : reg[0x%01x] = 0x%04x" % (val1,val4)
	reg[val1] = val4
	reg[0xf] += 3

def op_2_3(val1, val4):
	global reg
	print "OPCODE : reg[0x%01x] += 0x%04x" % (val1,val4)
	if val4 > val1 :
		reg[val1] = val4 - reg[val1]
	else :
		reg[val1] = reg[val1] - val4
	reg[0xf] += 3

def op_3_3(val1, val4):
	global reg
	print "OPCODE : reg[0x%01x] += 0x%04x" % (val1,val4)
	reg[val1] += val4
	reg[0xf] += 3

def op_4_3(val1, val4):
	global reg
	print "OPCODE : reg[0x%01x] << %d" % (val1,val4)
	reg[val1] = reg[val1] << val4
	reg[0xf] += 3

def op_5_3(val1, val4):
	global reg
	print "OPCODE : reg[0x%01x] >> %d" % (val1,val4)
	reg[val1] = reg[val1] >> val4
	reg[0xf] += 3

def op_7_3(val1, val4):
	global reg
	print "OPCODE : reg[0x%01x] = reg[0x%01x] & 0x%08x" % (val1,val1,val4)
	reg[val1] = reg[val1] & val4
	reg[0xf] += 3

def op_8_3(val4):
	global reg
	print "OPCODE : goto 0x%04x" % val4
	reg[0xf] = val4

def op_9_3(val1, val4):
	global reg
	print "OPCODE : goto 0x%04x if reg[0x%01x] != 0" % (val4,val1)
	if reg[val1] :
		reg[0xf] = val4
	else :
		reg[0xf] += 3

def op_0_0(val1, val2):
	global reg
	print "OPCODE : reg[0x%01x] = reg[0x%01x]" % (val1,val2)
	reg[val1] = reg[val2]
	reg[0xf] += 3

def op_1_0(val1):
	global reg
	print "OPCODE : reg[0x%01x]--" % (val1)
	reg[val1] = reg[val1] - 1
	reg[0xf] += 3

def op_2_0(val1, val2):
	global reg
	print "OPCODE : reg[0x%01x] += reg[0x%01x]" % (val1,val2)
	reg[val1] = reg[val1] | reg[val2]
	reg[0xf] += 3

def op_3_0(val1, val2):
	global reg
	print "OPCODE : reg[0x%01x] = abs(reg[0x%01x] - reg[0x%01x])" % (val1,val1,val2)
	if reg[val1] > reg[val2]:
		reg[val1] = reg[val1] - reg[val2]
	else :
		reg[val1] = reg[val2] - reg[val1]
	reg[0xf] += 3

def op_6_0(val1, val2):
	global reg
	print "OPCODE : reg[0x%01x] = reg[0x%01x] ^ reg[0x%01x]" % (val1,val1,val2)
	reg[val1] = reg[val1] ^ val2
	reg[0xf] += 3

def op_b_0(val1):
	global reg
	print "OPCODE : reg[0x%01x] = swapbyte16(reg[0x%01x])" % (val1,val1)
	reg[val1] = ((reg[val1] << 8) & 0xff00) |  ((reg[val1] >> 8) & 0xff)
	reg[0xf] += 3

# --------------------------------------------
#   Main
# --------------------------------------------

# Load VM code (SM4 decoded using r2)
filename = "./generated/vm_code_clear.r2"
for line in open(filename):
	(addr,code) = line.split(':')
	instr[int(addr,16)] = int(code.rstrip('\n'),16)

# Load lookup table
filename = "./generated/lookup_table.r2"
for line in open(filename):
	(val1,val2) = line.split(':')
	lookup_table[int(val1,16)] = int(val2.rstrip('\n'),16)

# Init the 16 reg
for i in range(16):
	reg[i] = 0

addr = 0
while global_return < 0 :
	process_one(addr, instr[addr])
	# for debug
	for i in range(0xf):
		reg[i] = 0
	# addr
	addr = addr + 3
