#! /usr/bin/python2

import r2pipe
import time
from os import system
import sys

# -------------------------------------
#                UTILS
# -------------------------------------

def launch_vm():
	# Spawn VM
	# My QEMU has been patched to use "-s" option by default
	# You may have to use a config file otherwise (as there is an antidebug protection)
	qemu_launch_cmd = 'qemu-system-aarch64 -nographic -machine virt,secure=on -cpu max -smp 1 -m 1024 -bios rom.bin -semihosting-config enable,target=native -device loader,file=./flash.bin,addr=0x04000000 -netdev user,id=network0,hostfwd=tcp:127.0.0.1:5555-192.168.200.200:22 -net nic,netdev=network0 -s -S 2>&1 > /dev/null &'
	qemu_kill_cmd = 'pkill -f qemu'
	system(qemu_kill_cmd)
	time.sleep(1)
	system(qemu_launch_cmd)
	time.sleep(2)

def launch_r2():
	# Spawn radare in TCP mode
	# NOTE VDE : unset LD_LIBRARY_PATH (not the same r2 in my path)
	r2_kill_cmd = 'pkill -f /usr/bin/radare2'
	system(r2_kill_cmd)
	time.sleep(1)
	r = r2pipe.open('gdb://127.0.0.1:1234', [ '-a',  'arm', '-b', '64', '-Dgdb', '-d', '-2'])
	time.sleep(2)
	return r

def launch_app(flag):
	# sleep 10 - Tunning to do
	system('bash -c \'sleep 60 ; sshpass -p sstic ssh -p 5555 root@127.0.0.1 ./safe_02/decrypted_file ' + flag + ' \' &')

def set_breakpoint(r, br):
	r.cmd('db 0x' + br)

def set_watchpoint(r, addr, mode):
	r.cmd('dbw 0x%08x %s' % (addr, mode))

def mem_to_128b_LE(hex):
	hex_128b_LE = bytearray('00000000000000000000000000000000')
	for i in range(len(hex)/2):
		hex_128b_LE[2*i] = hex[32-2*i-2]
		hex_128b_LE[2*i+1] = hex[32-2*i-1]
	return hex_128b_LE

def mem_to_64b_LE(hex):
	hex_64b_LE = bytearray('0000000000000000')
	for i in range(len(hex)/2):
		hex_64b_LE[2*i] = hex[16-2*i-2]
		hex_64b_LE[2*i+1] = hex[16-2*i-1]
	return hex_64b_LE

def mem_to_32b_LE(hex):
	hex_32b_LE = bytearray('00000000')
	for i in range(len(hex)/2):
		hex_32b_LE[2*i] = hex[8-2*i-2]
		hex_32b_LE[2*i+1] = hex[8-2*i-1]
	return hex_32b_LE

def get_32b_ptr(r, addr):
	return mem_to_32b_LE(r.cmd('p8 4 @ 0x%08x' % addr))

def get_64b_ptr(r, addr):
	return mem_to_64b_LE(r.cmd('p8 8 @ 0x%08x' % addr))

def get_128b_ptr(r, addr):
	return mem_to_128b_LE(r.cmd('p8 16 @ 0x%08x' % addr))

# -------------------------------------
#                NULL
# -------------------------------------

def null_handle(r):
	return None

def null_init(r):
	return null_handle

# -------------------------------------
#     DEBUG FUNCTIONS FOR SMC DUMP
# -------------------------------------

def breakpoint_handle_smc_dump(r):
	registers = r.cmdj('drj')
	#print "%08x" % registers['pc']
	if registers['pc'] == 0x0e037400: # Handler SMC
		#print "%08x" % registers['x0']
		#if registers['x0'] < 0xf0000000:
		print "%08x - %08x %08x %08x %08x %08x %08x" % (registers['x0'], registers['x1'], registers['x2'], registers['x3'], registers['x4'], registers['x5'], registers['x6'])

def smc_dump(r):
	set_breakpoint(r, '0e037400')
	return breakpoint_handle_smc_dump

# -------------------------------------
#       IOCTL1 DATA WATCHPOINT
# -------------------------------------

# The following dump the big buffer
# sent by userspace program during the
# first ioctl()
# -------------------------------------

def br_ioctl1_handle(r):
	registers = r.cmdj('drj')
	print "X0: %016x" % registers['x0']
	print "X1: %016x" % registers['x1']
	print "X2: %016x" % registers['x2']
	print "X19: %016x" % registers['x19']
	print "X22: %016x" % registers['x22']
	print r.cmd('px @ 0x%08x' % registers['x0'])

def init_ioctl1(r):
	set_breakpoint(r, '0e0310d0')
	return br_ioctl1_handle

# The following tries to retrieve where
# (which PC) we use the big buffer.
# -------------------------------------

watched = 0

def watchpoint_ioctl1_handle(r):
	global watched
	registers = r.cmdj('drj')
	if registers['x0'] == 0x83010004 and not watched :
		print "Setting watchpoint"
		r.cmd('db-*')
		set_watchpoint(r, 0x413000, 'r')
		set_watchpoint(r, 0x414000, 'r')
		watched=1
		#print r.cmd('db')
	if registers['pc'] != 0x0e037400 :
		print "PC: %08x" % registers['pc']

def watch_ioctl1_data(r):
	set_breakpoint(r, '0e037400')
	return watchpoint_ioctl1_handle

# The following tries to retrieve where
# (which PC) we use the big buffer NOT
# using watchpoint (because it does not
# work) by spying all registers.
# -------------------------------------

def manualwp_ioctl1_handle(r):
	global watched
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e037400: # Handler SMC
		print "%08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])
		#print "%08x" % registers['x0']
		if registers['x0'] == 0x83010004 :
			watched = 1
		if watched == 1:
			#code = r.cmdj('pij1')
			while registers['pc'] < 0x0f000000: # While in EL3 (not got back in kernel addr)
				r.cmd('ds')
				registers = r.cmdj('drj')
				for reg in range(30):
					if registers['x%d' % reg] >= 0x0e053000 and registers['x%d' % reg] < 0x0e154010 :
						print 'PC: 0x%08x' % registers['pc']
						return
				#code = r.cmdj('pdj1 @ 0x%08x' % registers['pc'])
				#print 'PC: 0x%08x' % registers['pc']
				#print code[0]['opcode']

def manualwp_ioctl1_init(r):
	set_breakpoint(r, '0e037400')
	return manualwp_ioctl1_handle

# -------------------------------------
#             MMAP DUMP
# -------------------------------------
# https://github.com/ARM-software/arm-trusted-firmware/blob/master/lib/xlat_tables_v2/xlat_tables_core.c#L653

def mmap_dump_handle(r):
	registers = r.cmdj('drj')
	print "New mapping [%016x] -> [%016x] size %d attr %08x" % (registers['x0'], registers['x1'], registers['x2'], registers['x3'])

def mmap_dump_init(r):
	set_breakpoint(r, '0e201af4')
	return mmap_dump_handle

# -------------------------------------
#     0e201a84 : wtf_function
# At startup :
# X0: 000000000e215a88
# X1: 0000000000000000
# X2: 000000000e215580
# X3: 000000000e21bfff
# X4: 00000000ffffffff
# X5: 0000000000000001
# -------------------------------------

def wtf1_handle(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e037400: # Handler SMC
		print "%08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])
	if registers['pc'] == 0x0e201a84:
		print "X0: %016x" % registers['x0']
		print "X1: %016x" % registers['x1']
		print "X2: %016x" % registers['x2']
		print "X3: %016x" % registers['x3']
		print "X4: %016x" % registers['x4']
		print "X5: %016x" % registers['x5']

def wtf1_init(r):
	set_breakpoint(r, '0e037400')
	set_breakpoint(r, '0e201a84')
	return wtf1_handle

# -------------------------------------
#       LAST SMC RETURNED PATH
# -------------------------------------

nth=0

def last_smc_handle(r):
	global nth
	registers = r.cmdj('drj')
	#print "%08x" % registers['pc']
	if registers['pc'] == 0x0e037400: # Handler SMC
		#print "%08x" % registers['x0']
		if registers['x0'] == 0xf2005002 :
			nth = nth + 1
			print "N=%d" % nth
			if nth == 17858:
				print "%08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])
				code = r.cmdj('pij1')
				while code[0]['opcode'] != 'eret':
					r.cmd('ds')
					registers = r.cmdj('drj')
					code = r.cmdj('pdj1 @ 0x%08x' % registers['pc'])
					print 'PC: 0x%08x' % registers['pc']
					if code[0]['opcode'] == 'eret' :
						print 'ERET'
					if code[0]['opcode'] == 'smc' :
						print 'SMC'

def last_smc_init(r):
	set_breakpoint(r, '0e037400')
	return last_smc_handle

# -------------------------------------
#       DEBUG STATE MACHINE F2XX
# -------------------------------------

mpidr_el1=-1

def debug_sm_f2xx(r):
	global mpidr_el1
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e037400:
		print ">>>>> %08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])

	if registers['pc'] == 0x0e200c08: # smc_f2xx_statemachine1 (start)
		print "sm_enter : %08x - %08x" % (registers['x0'], registers['x1'])

	if registers['pc'] == 0x0e200ca4: # calling f2xx_5002
		print "5002 : %016x" % registers['x0']

	if registers['pc'] == 0x0e200d10: # handling 5003
		print "5003 : index %d - %016x stored in %08x" % (registers['x3'], registers['x1'], registers['x2'])

	if registers['pc'] == 0x0e200c88: # handling 5001
		print "5001 : ... with value %08x" % registers['x0']

	if registers['pc'] == 0x0e200c70: # handling 5001
		mpidr_el1 = registers['x0']
		print "5001 : SMC 83010001 returned pointer %016x" % registers['x0']

	if registers['pc'] == 0x0e2022e4: # trigger_smc_f2001000
		print "f2001000 returned : %08x - %08x (stored into %08x)" % (registers['x0'], registers['x1'], registers['x4'])

	if registers['pc'] == 0x0e2022f4: # smc_f2xx_statemachine_atend -> print X0 = mpidr_el1
		mpidr_el1 = registers['x0']

	if registers['pc'] == 0x0e200cb8: # smc_f2xx_statemachine1 at end
		print "index %d (mpidr %08x) fID : %08x global2 : %08x global3 : %08x" % (registers['x2'], mpidr_el1, registers['x21'], registers['x20'], registers['x19'])
		mpidr_el1 = -1
		print "<<<<<"

def debug_sm_f2xx_init(r):
	set_breakpoint(r, '0e037400')
	set_breakpoint(r, '0e2022e4') # trigger_smc_f2001000
	set_breakpoint(r, '0e200c08') # smc_f2xx_statemachine1 (start)
	set_breakpoint(r, '0e200ca4') # calling f2xx_5002
	set_breakpoint(r, '0e200d10') # 5003
	set_breakpoint(r, '0e200c88') # 5001_1
	set_breakpoint(r, '0e200c70') # 5001_0
	set_breakpoint(r, '0e2022f4') # smc_f2xx_statemachine_atend -> print X0 = mpidr_el1
	set_breakpoint(r, '0e200cb8')
	return debug_sm_f2xx

def debug_sm_f2xx_5001(r):
	registers = r.cmdj('drj')
	print "SMC 83010001 returned %08x" % registers['x0']
	while registers['pc'] != 0x0e200c8c:
		r.cmd('ds')
		registers = r.cmdj('drj')
		code = r.cmdj('pdj1 @ 0x%08x' % registers['pc'])
		print 'PC: 0x%08x' % registers['pc']
		#print code[0]['opcode']
		sys.stdout.flush()
	sys.exit(1)

def debug_sm_f2xx_5001_init(r):
	set_breakpoint(r, '0e200c68')
	return debug_sm_f2xx_5001

# -------------------------------------
#   COUNT CMD 0x1100 on various input
# -------------------------------------

index=0
cmd = dict()

def count_cmd_11000(r):
	global cmd
	global index
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e2005c4:
		if registers['x0'] == 0x00011000:
			print "Index OK %d" % index
			cmd[index] = 1
			index = index + 1
		if registers['x0'] == 0x000d0001:
			cmd[index] = 0
			index = index + 1
		if registers['x0'] == 0x00a54000: # End
			print "End : " + repr(cmd)
			sys.stdout.flush()

def count_cmd_11000_init(r):
	set_breakpoint(r, '0e2005c4')
	return count_cmd_11000

# -------------------------------------
#            DEBUG 5002 CMDs
# -------------------------------------

def debug_states_5002(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e037400 and registers['x0'] == 0xf2005002:
		print ">>>> SMC %08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])
	if registers['pc'] == 0x0e037400 and registers['x0'] < 0xf2000000:
		print "SMC %08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])
	if registers['pc'] == 0x0e2005c4:
		print "x0 %08x cmd (w2) %02x" % (registers['x0'], registers['x2'])
		sys.stdout.flush()

def debug_states_5002_init(r):
	set_breakpoint(r, '0e037400')
	set_breakpoint(r, '0e2005c4')
	return debug_states_5002

# -------------------------------------
#  Tracing command 0xb smc 5002
# -------------------------------------

def trace_cmd_b_5002(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e2005c4 and registers['x2'] == 0xb:
		print "x0 %08x cmd (w2) %02x" % (registers['x0'], registers['x2'])
		while registers['pc'] < 0x0f000000: # While in EL3 (not got back in kernel addr)
			r.cmd('ds')
			registers = r.cmdj('drj')
			print 'PC: 0x%08x REGS %08x %08x %08x %08x' % (registers['pc'], registers['x0'], registers['x2'], registers['x3'], registers['x4'])
		sys.stdout.flush()

def trace_cmd_b_5002_init(r):
	set_breakpoint(r, '0e2005c4')
	return trace_cmd_b_5002

# -------------------------------------
#        Debug SMC trampoline
# -------------------------------------

trampoline_pc = 0
trampoline_sp = 0

def debug_trampoline1(r):
	global trampoline_pc
	global trampoline_sp
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e2025dc or registers['pc'] == 0x0e202664:
		trampoline_pc = registers['x4']
		trampoline_sp = registers['x6']
	if registers['pc'] == 0x0e20260c or registers['pc'] == 0x0e202694:
		while registers['pc'] == 0x0e20260c or registers['pc'] == 0x0e202694:
			r.cmd('ds')
			registers = r.cmdj('drj')
		print 'Trampoline pc=0x%08x : sp=0x%08x to : 0x%08x' % (trampoline_pc, trampoline_sp, registers['pc'])
		sys.stdout.flush()

def debug_trampoline1_init(r):
	set_breakpoint(r, '0e2025dc') # Trampoline 1
	set_breakpoint(r, '0e20260c') # Trampoline 1
	set_breakpoint(r, '0e202664') # Trampoline 2
	set_breakpoint(r, '0e202694') # Trampoline 2
	return debug_trampoline1

def debug_trampoline2(r):
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e2025dc or registers['pc'] == 0x0e202664:
		print 'MMU_ON  sp %08x X0 %08x X1 %08x' % (registers['sp'], registers['x0'], registers['x1'])
	if registers['pc'] == 0x0e20260c or registers['pc'] == 0x0e202694:
		print 'MMU_OFF sp %08x X0 %08x X1 %08x' % (registers['sp'], registers['x0'], registers['x1'])
	if registers['pc'] == 0x0e203604:
		print 'EXCPT   sp %08x X0 %08x X1 %08x' % (registers['sp'], registers['x0'], registers['x1'])
	if registers['pc'] == 0x0e202240:
		print 'SMC     sp %08x X0 %08x X1 %08x' % (registers['sp'], registers['x0'], registers['x1'])
		print '---'
		sys.stdout.flush()

def debug_trampoline2_init(r):
	set_breakpoint(r, '0e2025dc') # Trampoline 1
	set_breakpoint(r, '0e20260c') # Trampoline 1
	set_breakpoint(r, '0e202664') # Trampoline 2
	set_breakpoint(r, '0e202694') # Trampoline 2
	set_breakpoint(r, '0e203604') # Trampoline handler
	set_breakpoint(r, '0e202240') # Trampoline handler
	return debug_trampoline2

def debug_trampoline3(r):
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e203604 :
		print 'EXCPT   ' + mem_to_128b_LE(r.cmd('p8 16 @ 0x%08x' % 0x0e207240))
		sys.stdout.flush()
	if registers['pc'] == 0x0e20260c or registers['pc'] == 0x0e202694:
		print 'MMU_OFF ' + mem_to_128b_LE(r.cmd('p8 16 @ 0x%08x' % 0x0e207240))
		sys.stdout.flush()

def debug_trampoline3_init(r):
	set_breakpoint(r, '0e20260c') # Trampoline 1
	set_breakpoint(r, '0e202694') # Trampoline 2
	set_breakpoint(r, '0e203604') # Trampoline handler (beginning)
	return debug_trampoline3

def debug_trampoline4(r):
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e203604 :
		print 'TRAMP %016x%016x' % (registers['x1'], registers['x0'])
		sys.stdout.flush()
	if registers['pc'] == 0x0e202668 or registers['pc'] == 0x0e2025e0:
		print '5002 SPSR %08x' % registers['x6']
		sys.stdout.flush()

def debug_trampoline4_init(r):
	set_breakpoint(r, '0e2025e0') # Trampoline 1
	set_breakpoint(r, '0e202668') # Trampoline 2
	set_breakpoint(r, '0e203604') # Trampoline handler (beginning)
	return debug_trampoline4

def debug_trampoline5(r):
	global trampoline_pc
	global trampoline_sp
	registers = r.cmdj('drj')

	if registers['pc'] == 0e202694:
		# DONT WORK : r.cmd('dr x6 = 0x0d0') # Enable debugging
		while registers['pc'] != 0x0e203604:
			r.cmd('ds')
			registers = r.cmdj('drj')
			code = r.cmdj('pdj1 @ 0x%08x' % 0x0e205230)
			print code[0]['opcode']
			print 'pc=0x%08x - %08x %08x %08x %08x %08x %08x %08x' \
				% (registers['pc'], registers['x0'], registers['x1'],
				   registers['x2'], registers['x3'], registers['x4'],
				   registers['x5'], registers['x6'])
			sys.stdout.flush()
		sys.exit(1)

def debug_trampoline5_init(r):
	set_breakpoint(r, '0e202694') # Trampoline 1
	return debug_trampoline5

start_logging=0
def debug_trampoline_cmd09(r):
	global start_logging
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e037400 and registers['x0'] < 0xf2000000:
		if start_logging :
			print "SMC %08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])

	if registers['pc'] == 0x0e031224:
		if start_logging :
			print "SMC 83010001 returned %08x" % registers['x0']

	if registers['pc'] == 0x0e2005c4:
		if ((registers['x0'] >> 20) == 0x9) and (((registers['x0'] >> 18) & 3) == 3):
			start_logging = 1
		if start_logging :
			print "CMD %08x" % registers['x0']

	if registers['pc'] == 0x0e202664:
		if start_logging :
			print 'SPSR : %08x' % registers['x6']
			print 'ELR  : %08x' % registers['x4']

	if registers['pc'] == 0x0e200b5c:
		print "----"
		start_logging = 0
		sys.stdout.flush()

def debug_trampoline_cmd09_init(r):
	set_breakpoint(r, '0e037400')
	set_breakpoint(r, '0e2005c4')
	set_breakpoint(r, '0e031224')
	set_breakpoint(r, '0e202664') # Trampoline 2
	set_breakpoint(r, '0e200b5c') # Trampoline 2
	return debug_trampoline_cmd09

# -------------------------------------
#         Debug F2xx5002 NEON
# -------------------------------------

def debug_5002_NEON(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e2009a0:
		while registers['pc'] <= 0x0f000000:
			r.cmd('ds')
			registers = r.cmdj('drj')
			code = r.cmdj('pdj1 @ 0x%08x' % registers['pc'])
			print 'PC: 0x%08x' % registers['pc']
			print code[0]['opcode']
			sys.stdout.flush()
		sys.exit(1)

def debug_5002_NEON_init(r):
	set_breakpoint(r, '0e2009a0') # cmd20 = 2
	return debug_5002_NEON


# ---------------------------------------------------
#  Check if other entries on vector F2xx are called
# ---------------------------------------------------

vect_base = 0x0e200090

def debug_vector(r):
	global vect_base
	registers = r.cmdj('drj')
	print 'Vector entry %d' % ((registers['pc'] - vect_base) / 4)
	sys.stdout.flush()

def debug_vector_init(r):
	global vect_base
	for i in range(10):
		set_breakpoint(r, '%08x' % (vect_base + i*4))
	return debug_vector

# Result : vector 1 and 6 are used
# Vector 6 tick periodically from startup

# ----------------------------
#  Debug pagefault exec flow
# ----------------------------

# Page fault seen at
# * SMC F2xx5003 : 0e200d0c x1 -> x0
# * SMC F2xx5001 : 0e200c7c x0 -> x0
# Handler at 0e203200 (+ a few instructions)
# Custom function called at 0e20228c args = x0, x1
# New value for elr_el1 at 0e20229c (value+4)

cmd_encoded = 0

def debug_pf0(r):
	global cmd_encoded
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e200d0c:
		print 'SMC F2xx5003 X1 %08x -> X0 %08x' % (registers['x1'], registers['x0'])
	if registers['pc'] == 0x0e200c7c:
		print 'SMC F2xx5001 X0 %08x -> X0 %08x' % (registers['x0'], registers['x0'])
	if registers['pc'] == 0x0e2005a4:
		cmd_encoded = registers['x0']
	if registers['pc'] == 0x0e200b94 or registers['pc'] == 0x0e200624:
		print 'SMC F2xx5002 (cmd = %08x) X0 %08x -> X0 %08x' % (cmd_encoded, registers['x0'], registers['x0'])
	if registers['pc'] == 0x0e20228c:
		print 'STORE_RESULT(%08x , %08x)' % (registers['x0'], registers['x1'])
	if registers['pc'] == 0x0e20229c:
		print 'elr_el1 = X0 %08x' % registers['x0']
	sys.stdout.flush()

def debug_pf0_init(r):
	set_breakpoint(r, '0e200d0c') # SMC F2xx5003
	set_breakpoint(r, '0e200c7c') # SMC F2xx5001
	set_breakpoint(r, '0e2005a4') # SMC F2xx5002 begin
	set_breakpoint(r, '0e200b94') # SMC F2xx5002 case 1
	set_breakpoint(r, '0e200624') # SMC F2xx5002 case 2
	set_breakpoint(r, '0e20228c') # store_result()
	set_breakpoint(r, '0e20229c') # elr_el1 -> go back to f2xx handler after next ERET
	return debug_pf0

# -------------------------------------
#      STORE RESULT ON PAGEFAULT
# -------------------------------------

def store_result_handle_bak(r):
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e200e84: # Function begin
		print "store_result(0x%08x, 0x%08x)" % (registers['x0'], registers['x1'])
		print r.cmd('pxl 4 @ 0x%08x' % registers['x0'])

	if registers['pc'] == 0x0e200eb4: # Memcpy
		print "memcpy(0x%08x, 0x%08x, 0x10)" % (registers['x0'], registers['x1'])
		print r.cmd('pxl 1 @ 0x%08x' % registers['x0'])

def store_result_init_bak(r):
	set_breakpoint(r, '0e200e84')
	set_breakpoint(r, '0e200eb4')
	return store_result_handle_bak

faulted_value = 0
bigbuf_in_addr = 0
bigbuf_in_hex = '00000000000000000000000000000000'
bigbuf_in2_hex = '00000000000000000000000000000000'
res1_hex = '00000000000000000000000000000000'
res2_hex = '00000000000000000000000000000000'
res3_hex = '00000000000000000000000000000000'
res4_hex = '00000000000000000000000000000000'
new_x0 = 0
old_x1 = 0
old_x2 = 0

def store_result_handle(r):
	global faulted_value
	global bigbuf_in_addr
	global bigbuf_in_hex
	global bigbuf_in2_hex
	global res1_hex
	global res2_hex
	global res3_hex
	global res4_hex
	global new_x0
	global old_x1
	global old_x2
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e200770: # Always "nop" antidebug instr
		r.cmd('dr x2 = 0xd') # Just do pc = pc + 3
	if registers['pc'] == 0x0e200eac: # Here IN4 is in x21 (faulted_value)
		faulted_value = registers['x21']
	if registers['pc'] == 0x0e200f20: # BIGBUF[IN4]
		bigbuf_in_addr = registers['x0']
		bigbuf_in_hex = r.cmd('p8 16 @ 0x%08x' % bigbuf_in_addr)
	if registers['pc'] == 0x0e200f90: # BIGBUF[IN4+0x10]
		bigbuf_in2_hex = r.cmd('p8 16 @ 0x%08x' % registers['x2'])
	if registers['pc'] == 0x0e200f40: # res1
		res1_hex = r.cmd('p8 16 @ 0x%08x' % registers['x20'])
	if registers['pc'] == 0x0e200fb0: # res2
		if (faulted_value & 0xf) > 0xc:
			res2_hex = r.cmd('p8 16 @ 0x%08x' % registers['x3'])
	if registers['pc'] == 0x0e201010: # res3
		res3_hex = r.cmd('p8 16 @ 0x%08x' % registers['x0'])
	if registers['pc'] == 0x0e201078: # res4
		res4_hex = r.cmd('p8 16 @ 0x%08x' % registers['x0'])
	if registers['pc'] == 0x0e20108c: # new_x0
		new_x0 = registers['x0']
	if registers['pc'] == 0x0e200fc4: # save x1 in RAM
		old_x1 = registers['x2']
	if registers['pc'] == 0x0e200fb8: # x2 == 1 -> special processing
		old_x2 = registers['x2']
	if registers['pc'] == 0x0e2010a8: # End of the function, print all data
		print '1 %016x %016x %016x %016x %016x ' % (faulted_value, old_x2, old_x1, new_x0, bigbuf_in_addr)
		print '2 ' + bigbuf_in_hex + ' ' + res1_hex + ' ' + res3_hex + ' ' + bigbuf_in2_hex + ' ' + res2_hex + ' ' + res4_hex
		sys.stdout.flush()
		faulted_value = 0
		bigbuf_in_addr = 0
		bigbuf_in_hex = '00000000000000000000000000000000'
		bigbuf_in2_hex = '00000000000000000000000000000000'
		res1_hex = '00000000000000000000000000000000'
		res2_hex = '00000000000000000000000000000000'
		res3_hex = '00000000000000000000000000000000'
		res4_hex = '00000000000000000000000000000000'
		new_x0 = 0
		old_x1 = 0
		old_x2 = 0

def store_result_init(r):
	set_breakpoint(r, '0e200770') # Bypass antidebug
	set_breakpoint(r, '0e200eac')
	set_breakpoint(r, '0e200f20')
	set_breakpoint(r, '0e200f40')
	set_breakpoint(r, '0e200f90')
	set_breakpoint(r, '0e200fb0')
	set_breakpoint(r, '0e201010')
	set_breakpoint(r, '0e201078')
	set_breakpoint(r, '0e20108c')
	set_breakpoint(r, '0e200fc4')
	set_breakpoint(r, '0e200fb8')
	set_breakpoint(r, '0e2010a8')
	return store_result_handle

# ---------------------------------------
# DEBUG FUNCTIONS FOR 0x83010002 tracing
# ---------------------------------------

def breakpoint_handle_83010002(r):
	registers = r.cmdj('drj')
	#print "%08x" % registers['pc']
	if registers['pc'] == 0x0e037400: # Handler SMC
		#print "%08x" % registers['x0']
		if registers['x0'] == 0x83010002 :
			print "%08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])
			code = r.cmdj('pij1')
			while code[0]['opcode'] != 'eret':
				r.cmd('ds')
				registers = r.cmdj('drj')
				code = r.cmdj('pdj1 @ 0x%08x' % registers['pc'])
				#print 'PC: 0x%08x' % registers['pc']
				#print code[0]['opcode']

				# MEMCPY()
				if registers['pc'] == 0x0e036070:
					print "uvar8 (x19) = %08x" % registers['x19']
					print "memcpy(dest=%08x,source=%08x,size=%08x)" % (registers['x0'], registers['x1'], registers['x2'])
					print r.cmd('pxl 1 @ 0x%08x' % registers['x1'])
				# AES KEY PART 1
				if registers['pc'] == 0x0e031270:
					print "key[0] = %08x" % (registers['x0'])
				# AES KEY PART 2
				if registers['pc'] == 0x0e031278:
					print "key[1] = %08x" % (registers['x0'])
				# AES KEY PART 3
				if registers['pc'] == 0x0e031280:
					print "key[2] = %08x" % (registers['x0'])
				# AES KEY PART 4
				if registers['pc'] == 0x0e031288:
					print "key[3] = %08x" % (registers['x0'])
				# AES DECRYPTION
				if registers['pc'] == 0x0e03128c:
					print "aesd(data = %08x)" % (registers['x20'])
					print r.cmd('pxl 1 @ 0x%08x' % registers['x20'])
				# AES+XOR Result
				if registers['pc'] == 0x0e03129c:
					print r.cmd('pxl 1 @ 0x%08x' % registers['x19'])
				# Function returns
				if registers['pc'] == 0x0e031150:
					print "end of SMC_8301000x_handler()"
					print "X0  = %08x)" % (registers['x0'])
					print "X19 = %08x)" % (registers['x19'])
					print "X20 = %08x)" % (registers['x20'])
					print "X21 = %08x)" % (registers['x21'])
					print "X22 = %08x)" % (registers['x22'])
					print "X23 = %08x)" % (registers['x23'])
					print "X24 = %08x)" % (registers['x24'])
					print "X29 = %08x)" % (registers['x29'])
					print "X30 = %08x)" % (registers['x30'])

				if code[0]['opcode'] == 'eret' :
					print 'ERET'
				if code[0]['opcode'] == 'smc' :
					print 'SMC'
			#sys.exit(1)

def trace_83010002(r):
	set_breakpoint(r, '0e037400')
	return breakpoint_handle_83010002

# ---------------------------------------
# DEBUG FUNCTIONS FOR 0x83010001 tracing
# ---------------------------------------

fpexc32_el2 = 0
dacr32_el2 = 0
ifsr32_el2 = 0
sder32_el3 = 0
operand1_hex = '00000000000000000000000000000000'
result_hex = '00000000000000000000000000000000'

def breakpoint_handle_83010001(r):
	global fpexc32_el2
	global dacr32_el2
	global ifsr32_el2
	global sder32_el3
	global operand1_hex
	global result_hex

	registers = r.cmdj('drj')
	#print "%08x" % registers['pc']

	if registers['pc'] == 0x0e0311cc: # Start of case 80001
		return
	if registers['pc'] == 0x0e0311e0: # x19 -> operand1
		operand1_hex = mem_to_128b_LE(r.cmd('p8 16 @ 0x%08x' % registers['x19']))
	if registers['pc'] == 0x0e0311e4: # x0 = fpexc32_el2
		fpexc32_el2 = registers['x0']
	if registers['pc'] == 0x0e0311ec: # x0 = dacr32_el2
		dacr32_el2 = registers['x0']
	if registers['pc'] == 0x0e0311f4: # x0 = ifsr32_el2
		ifsr32_el2 = registers['x0']
	if registers['pc'] == 0x0e0311fc: # x0 = sder32_el3
		sder32_el3 = registers['x0']
	if registers['pc'] == 0x0e031210: # x1 -> result
		result_hex = mem_to_128b_LE(r.cmd('p8 16 @ 0x%08x' % registers['x1']))
	if registers['pc'] == 0x0e031224: # end
		print '80001 ret %08x AESE %08x%08x%08x%08x ' % (registers['x0'], sder32_el3, ifsr32_el2, dacr32_el2, fpexc32_el2) + operand1_hex + ' -> ' + result_hex
		sys.stdout.flush()
		fpexc32_el2 = 0
		dacr32_el2 = 0
		ifsr32_el2 = 0
		sder32_el3 = 0
		operand1_hex = '00000000000000000000000000000000'
		result_hex = '00000000000000000000000000000000'

def trace_83010001(r):
	set_breakpoint(r, '0e0311cc')
	set_breakpoint(r, '0e0311e0')
	set_breakpoint(r, '0e0311e4')
	set_breakpoint(r, '0e0311ec')
	set_breakpoint(r, '0e0311f4')
	set_breakpoint(r, '0e0311fc')
	set_breakpoint(r, '0e031210')
	set_breakpoint(r, '0e031224')
	return breakpoint_handle_83010001

# ---------------------------------------
# 83010011
# ---------------------------------------

# There is some dead code in the handler of SMC 8301000xx
# As this looks suspect, this test checks that the address
# is never accessed.
def debug_830100xx_check_dead_code(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e03187c:
		print 'Dead code 0x0e03187c is reached!'
		sys.stdout.flush()
	if registers['pc'] == 0x0e03187c:
		print 'Dead code 0x0e0318a4 is reached!'
		sys.stdout.flush()
	if registers['pc'] == 0x0e031870:
		print 'Limit case is reached'
		sys.stdout.flush()

def debug_830100xx_check_dead_code_init(r):
	set_breakpoint(r, '0e03187c')
	set_breakpoint(r, '0e0318a4')
	set_breakpoint(r, '0e031870')
	return debug_830100xx_check_dead_code

# Result : code is reached... There must be some trick like VFP exception...
# Code seems to really be the handling of 83010011 but there is an unknown instruction "setffr" in Ghisdra

def trace_83010011(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e0317f4:
		while registers['pc'] != 0x0e031998:
			r.cmd('ds')
			registers = r.cmdj('drj')
			code = r.cmdj('pdj1 @ 0x%08x' % registers['pc'])
			print code[0]['opcode']
			print 'pc=0x%08x - %08x %08x %08x %08x %08x %08x %08x' \
				% (registers['pc'], registers['x0'], registers['x1'],
				   registers['x2'], registers['x3'], registers['x4'],
				   registers['x5'], registers['x6'])
			sys.stdout.flush()
		sys.exit(1)

def trace_83010011_init(r):
	set_breakpoint(r, '0e0317f4')
	return trace_83010011

# --------------------------------------------
# 83010022
# --------------------------------------------
in1 = 0
in2 = 0
out = 0

def trace_83010022(r):
	global in1
	global in2
	global out
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e031bc0:
		in1 = registers['x1']
		in2 = registers['x22']
	if registers['pc'] == 0x0e031be4:
		out = registers['x22']
		print "%08x + %08x = %08x" % (in1, in2, out)
		sys.stdout.flush()

def trace_83010022_init(r):
	set_breakpoint(r, '0e031bc0')
	set_breakpoint(r, '0e031be4')
	return trace_83010022

# --------------------------------------------
# 83010012
# --------------------------------------------

# Reversing 83010012 shows :
# 1. payload[arg1] |= payload[arg2] (without overlap)
# 2. payload[0xf] += 3

def debug_83010012_payloads(r):
	global operand1_fcadd
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e0314c4:
		print 'payload[0xf] %08x -> %08x' % (registers['x0'], registers['x0']+3)
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e031400:
		operand1_fcadd = registers['x20']
	if registers['pc'] == 0x0e031408:
		print 'payload[0x%x] %08x -> %08x' % (registers['x19'], operand1_fcadd, registers['x0'])
		sys.stdout.flush()

def debug_83010012_payloads_init(r):
	set_breakpoint(r, '0e0314c4')
	set_breakpoint(r, '0e031400')
	set_breakpoint(r, '0e031408')
	return debug_83010012_payloads

# --------------------------------------------
# 83010027
# --------------------------------------------

# Reversing 83010012 shows :
# 1. payload[arg1] |= payload[arg2] (without overlap)
# 2. payload[0xf] += 3

def debug_83010027_payloads(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e031e90:
		print get_64b_ptr(r, registers['x0']) + ' & ' + get_64b_ptr(r, registers['x23']) + ' -> ' + get_64b_ptr(r, registers['x20'])
		sys.stdout.flush()

def debug_83010027_payloads_init(r):
	set_breakpoint(r, '0e031e90')
	return debug_83010027_payloads

# --------------------------------------------
# DEBUG FUNCTIONS FOR UNDERSTAND FCADD INSTR
# --------------------------------------------

operand1_fcadd=0
operand2_fcadd=0

def debug_83010012_fcadd(r):
	global operand1_fcadd
	global operand2_fcadd
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e031400:
		operand1_fcadd = registers['x20']
		operand2_fcadd = registers['x0']
	if registers['pc'] == 0x0e031408:
		print 'FCADD %08x + %08x (+rot) -> %016x' % (operand1_fcadd, operand2_fcadd, registers['x0'])
		sys.stdout.flush()

# Result :
# * this acts as a normal interger addition
# * no data loss as op1 and op2 are never on the same bits
# Ex : FCADD 0006d400 + 00000094 (+rot) -> 000000000006d494

def debug_83010012_fcadd_init(r):
	set_breakpoint(r, '0e031400')
	set_breakpoint(r, '0e031408')
	return debug_83010012_fcadd

def debug_83010013_fcadd(r):
	global operand1_fcadd
	global operand2_fcadd
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e0315ec:
		operand1_fcadd = registers['x23']
		operand2_fcadd = registers['x0']
	if registers['pc'] == 0x0e0315f4:
		print 'FCADD %08x + %08x -> %016x' % (operand1_fcadd, operand2_fcadd, registers['x0'])
		sys.stdout.flush()

def debug_83010013_fcadd_init(r):
	set_breakpoint(r, '0e0315ec')
	set_breakpoint(r, '0e0315f4')
	return debug_83010013_fcadd

def debug_83010011_fcadd(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e0318a8:
		print 'At 0e0318a4 : tmp16 = %016x' % registers['x0']
		sys.stdout.flush()
	if registers['pc'] == 0x0e0318b8:
		print 'tmp16 + 1 = %016x' % registers['x0']
	if registers['pc'] == 0x0e0318cc:
		print 'tmp4 = %08x - tmp16 = %08x' % (registers['x23'], registers['x0'])
		sys.stdout.flush()

def debug_83010011_fcadd_init(r):
	set_breakpoint(r, '0e0318a8')
	set_breakpoint(r, '0e0318b8')
	set_breakpoint(r, '0e0318cc')

	return debug_83010011_fcadd

def debug_83010011_fcadd2(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e031898:
		print 'TMP6 = ' + get_128b_ptr(r, registers['x20'])
		sys.stdout.flush()

def debug_83010011_fcadd2_init(r):
	set_breakpoint(r, '0e031898')
	return debug_83010011_fcadd2


# --------------------------------------------
# UNKNOWN INSTRUCTION RUNTIME DEBUG
# --------------------------------------------

def debug_83010016_ukwn(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e0316dc:
		print 'TMP1 %016x TMP2 %016x' % (registers['x24'],registers['x0'])
	if registers['pc'] == 0x0e0316f0:
		print 'IN1 ' +  get_64b_ptr(r, registers['x0']) + ' IN2 ' + get_64b_ptr(r, registers['x23'])
	if registers['pc'] == 0x0e03170c:
		print 'OUT ' + get_64b_ptr(r, registers['x20'])
		sys.stdout.flush()

def debug_83010016_ukwn_init(r):
	set_breakpoint(r, '0e0316dc')
	set_breakpoint(r, '0e03170c')
	set_breakpoint(r, '0e0316f0')

	return debug_83010016_ukwn

# -------------------------------------
#    cRead VM memory
# -------------------------------------

VM_memory=0x1000
pf_to_handle=0
incr=4

def dump_vm_code(r):
	global VM_memory
	global pf_to_handle
	global incr
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e200c78: # print pagefault result (not the first time)
		# Inject value
		if VM_memory > 0x1000:
			print "%08x : %08x" % (VM_memory-incr, registers['x0'] & 0xffffffff)
			sys.stdout.flush()
		if VM_memory > 0x101010: # End of VM code
			sys.exit(0)

	if registers['pc'] == 0x0e200c7c: # override pagefault input
		r.cmd('dr x0 = 0x%08x' % VM_memory)
		pf_to_handle = 1

	if registers['pc'] == 0x0e20229c: # go back to 1
		if pf_to_handle == 1 :
			r.cmd('dr x0 = 0x%08x' % 0x0e200c78)
			VM_memory = VM_memory + incr
		pf_to_handle = 0

def dump_vm_code_init(r):
	set_breakpoint(r, '0e200c78') # before page fault
	set_breakpoint(r, '0e200c7c') # before page fault
	set_breakpoint(r, '0e20229c') # at the end of the pagefault, go back to first breakpoint
	return dump_vm_code

# ---------------------------------------
# DEBUG FUNCTIONS FOR VM CMD18 = 1
# ---------------------------------------

faulted_value=0

def debug_cmd_pf(r):
	global start_logging
	global faulted_value
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e2005c4:
		cmd = registers['x0']
		if ((cmd >> 20) & 0xff) >= 0xc:
			return
		if ((cmd >> 18) & 0x3) == 0x1 :
			start_logging = 1
			print "CMD %08x" % registers['x0']

	if registers['pc'] == 0x0e200b94:
		faulted_value = registers['x0']

	if registers['pc'] == 0x0e200b98:
		print "Fault %08x -> %08x" % (faulted_value, registers['x0'])
		faulted_value = 0

	if registers['pc'] == 0x0e037400 and registers['x0'] < 0xf2000000:
		if start_logging :
			print "SMC %08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])

	if registers['pc'] == 0x0e031224:
		if start_logging :
			print "SMC 83010001 returned %08x" % registers['x0']

	if registers['pc'] == 0x0e200bb4:
		start_logging = 0
		print "----"
		sys.stdout.flush()

def debug_cmd_pf_init(r):
	set_breakpoint(r, '0e2005c4') # Cmd parsing
	set_breakpoint(r, '0e200b94') # Before PF
	set_breakpoint(r, '0e200b98') # After PF
	set_breakpoint(r, '0e200bb4') # End
	set_breakpoint(r, '0e037400') # SMC 83xx enter
	set_breakpoint(r, '0e031224') # SMC 83xx end
	return debug_cmd_pf

# --------------------------------------------------
# DEBUG FUNCTIONS FOR VM CMD18 = 2 and CMD20 = 0xc
# --------------------------------------------------

new_value = 0

def debug_cmd_pf2(r):
	global start_logging
	global faulted_value
	global new_value
	registers = r.cmdj('drj')

	if registers['pc'] == 0x0e2005c4:
		cmd = registers['x0']
		if ((cmd >> 20) & 0xff) > 0xc:
			return
		if ((cmd >> 20) & 0xff) == 0xc:
			start_logging = 1
			print "CMD20_C %08x" % registers['x0']
		if ((cmd >> 18) & 0x3) == 0x2 :
			start_logging = 1
			print "CMD18_2 %08x" % registers['x0']

	if registers['pc'] == 0x0e200610:
		print "CMD20_C Jumping at %08x" % registers['x0']

	if registers['pc'] == 0x0e200624:
		if registers['x2'] != 1 :
			print "Error: Here is not only a storage"
		faulted_value = registers['x0']
		new_value = registers['x1']

	if registers['pc'] == 0x0e200628:
		if start_logging :
			print "Fault. New val %08x at idx %08x" % (new_value, faulted_value)
			print "----"
		faulted_value = 0
		start_logging = 0
		sys.stdout.flush()

	if registers['pc'] == 0x0e037400 and registers['x0'] < 0xf2000000:
		if start_logging :
			print "SMC %08x - %08x %08x" % (registers['x0'], registers['x1'], registers['x2'])

	if registers['pc'] == 0x0e031224:
		if start_logging :
			print "SMC 83010001 returned %08x" % registers['x0']

def debug_cmd_pf2_init(r):
	set_breakpoint(r, '0e2005c4') # Cmd parsing
	set_breakpoint(r, '0e200624') # Before PF
	set_breakpoint(r, '0e200628') # After PF
	set_breakpoint(r, '0e037400') # SMC 83xx enter
	set_breakpoint(r, '0e031224') # SMC 83xx end
	set_breakpoint(r, '0e200610') # Jump for cmd 20
	return debug_cmd_pf2

# -------------------------------------
#     DEBUG POINTER AT 0x0x09010000
# -------------------------------------

def check_ptr_09(r):
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e200178: # Handler SMC
		print "%08x - %08x %08x - " % (registers['x0'], registers['x1'], registers['x2']) + get_32b_ptr(r, 0x09010000) + ' ' + get_32b_ptr(r, 0x09010008)

def check_ptr_09_init(r):
	set_breakpoint(r, '0e200178')
	return check_ptr_09

# -------------------------------------
#  MONITOR CHANGES ON REGISTERS (VM)
# -------------------------------------

get_val_idx = 0

def monitor_reg_changes(r):
	global get_val_idx
	registers = r.cmdj('drj')
	if registers['pc'] == 0x0e200770: # Always "nop" antidebug instr
		r.cmd('dr x2 = 0xd') # Just do pc = pc + 3
	if registers['pc'] == 0x0e037400: # Handler SMC
		if registers['x0'] < 0xf0000000:
			if registers['x0'] == 0x83010001:
				get_val_idx = registers['x1']
			else :
				print "SET (op %02x) - %08x %08x" % (registers['x0'] & 0xff, registers['x1'], registers['x2'])
				sys.stdout.flush()
	if registers['pc'] == 0x0e031224:
		print "GET %08x = %08x" % (get_val_idx, registers['x0'])
		sys.stdout.flush()

def monitor_reg_changes_init(r):
	set_breakpoint(r, '0e037400')
	set_breakpoint(r, '0e031224')
	set_breakpoint(r, '0e200770') # bypass anti-debug
	return monitor_reg_changes

# -------------------------------------
#                MAIN
# -------------------------------------

print 'Launching QEMU...'
launch_vm()

print 'Launching Radare2...'
r = launch_r2()

print 'Initialization...'
br_handle = store_result_init(r)

print r.cmd('db')

print 'Scheduling application...'
#launch_app('AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBCCCCCCCCCCCCCCCCDDDDDDDDDDDDDDDD')
#launch_app('53535449437B787878787878787878787878787878787878787878787878787D')
#launch_app('0da5866a1b9d2dd37386efa4e5ff76ae087044465f22d7eb911aaa89ec260e74')
launch_app('acadaa8b5b55306fb3c6dfc3b2d1c80770084644225febd71a9189aa26ec740e')
#launch_app('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA')
#launch_app('612e72706766722e666e632e2e76662e76706e7366407279707667667465622e')
#launch_app('76706e7366407279707667667465622e612e72706766722e666e632e2e76662e')
#launch_app('2e76662e666e632e6766722e612e72707465622e707667666640727976706e73')

print 'Running...'
while True:
	r.cmd('dc')
	br_handle(r)

