#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>

#include "lookup_table.h"
unsigned char* lookup_table = (unsigned char*) lookup_table_int;

#define decrease_j(j) { if (j == 0) { j = 0xa; } j--; }

static inline uint16_t lookup(unsigned int j, uint16_t in1)
{
	unsigned char r4, r5, r6;
	uint32_t idx;
	uint16_t in1_l, in1_h;

	in1_l = in1 & 0xff;
	in1_h = (in1 >> 8) & 0xff;

	idx = (j << 16) + (in1_l << 8) + in1_h;
	r6 = lookup_table[idx] & 0xff;
	//printf("idx %08x val %02x j %d\n", idx + 0x1000, r6, j);

	decrease_j(j)

	idx = (j << 16) + (in1_h << 8) + r6;
	r5 = lookup_table[idx] & 0xff;
	//printf("idx %08x val %02x j %d\n", idx + 0x1000, r5, j);

	decrease_j(j)

	idx = (j << 16) + (r6 << 8) + r5;
	r4 = lookup_table[idx] & 0xff;
	//printf("idx %08x val %02x j %d\n", idx + 0x1000, r4, j);

	decrease_j(j)

	idx = (j << 16) + (r5 << 8) + r4;
	r6 = lookup_table[idx] & 0xff;
	//printf("idx %08x val %02x j %d\n", idx + 0x1000, r6, j);

	decrease_j(j)

	return (r6 << 8) | r4;
}

static uint16_t lookup_reverse(int j, uint16_t r9)
{
	uint32_t try = 0;
	uint16_t result = 0;
	uint16_t good = 0;
	int found = 0;

	j = (j + 4) % 10;

	/* We just have 2^16 possibility, so let's bruteforce... */
	do {
		result = lookup(j, try);
		if (result == r9) {
			//printf("Decoded 0x%04x (j=%d) : 0x%04x\n", r9, j, try);
			good = try;
			found = 1;
		}
		try++;
	} while (try < 0x10000);

	if (!found) {
		printf("Failed to decode 0x%04x (j=%d)\n", r9, j);
	}
	return good;
}

void self_test1() {
	// Value extracted from SMC parameters
	// Lookup 7baaa -> 0xee
	printf("0x%02x == 0xee\n", lookup_table[0x7baaa-0x1000]);
	printf("0x%02x == 0x47\n", lookup_table[0x1cf67-0x1000]);

}

void self_test2() {
	// Value extracted from SMC parameters
	//printf("0x%04x == 0x9713\n", lookup_reverse(9, 0x4173));
	//printf("0x%04x == 0x4173\n", lookup(3, 0x9713));
	printf("0x%04x == 0xb4c5\n", lookup(1, 0x67bf));
}

void self_test3() {
	// Value extracted from SMC parameters
	//printf("0x%04x == 0x67bf\n", lookup(1, 0xb4c5));
}


void reverse(uint16_t in0, uint16_t in1, uint16_t in2, uint16_t in3)
{
	int i = 0;
	int j = 9; // Final value of j (seen in trace generated by r2)
	uint32_t r8;

	for (i=0x00; i<0x20; i++) {
		if (((i >> 3) & 1) == 0) {
			//printf("Algo 1\n");
			r8 = in2;
			in2 = in1;
			in1 = lookup_reverse(j, in0);
			j = (j + 4) % 10;
			in0 = in3 ^ in1 ^ (i+1);
			in3 = r8;
		} else { // Ox1f -> 0x18, 0xf -> 0x8
			//printf("Algo 2\n");
			r8 = in3;
			in3 = in2;
			in2 = in1 ^ in0 ^ (i+1);
			in1 = lookup_reverse(j, in0);
			j = (j + 4) % 10;
			in0 = r8;
		}
		//printf("New values : 0x%04x 0x%04x 0x%04x 0x%04x\n", in0, in1, in2, in3);
	}
	printf("Result : 0x%04x 0x%04x 0x%04x 0x%04x\n", in0, in1, in2, in3);
}

int main(int argc, char** argv)
{
	uint16_t in0, in1, in2, in3;

	//self_test1();
	//self_test2();
	//self_test3();

	/* Expected output for IN=0xaaaaaaaaaaaaaaaa */
	in0 = 0x4173;
	in1 = 0x9333;
	in2 = 0xe409;
	in3 = 0x8441;
	reverse(in0, in1, in2, in3);

	/* Real output for good input */
	in0 = 0x7072;
	in1 = 0x2e61;
	in2 = 0x2e72;
	in3 = 0x6667;
	reverse(in0, in1, in2, in3);

	in0 = 0x2e63;
	in1 = 0x6e66;
	in2 = 0x2e66;
	in3 = 0x762e;
	reverse(in0, in1, in2, in3);

	in0 = 0x736e;
	in1 = 0x7076;
	in2 = 0x7972;
	in3 = 0x4066;
	reverse(in0, in1, in2, in3);

	in0 = 0x6667;
	in1 = 0x7670;
	in2 = 0x2e62;
	in3 = 0x6574;
	reverse(in0, in1, in2, in3);

	return 0;
}
